/*
 * sm.h
 *
 */

#if !defined(_SM_INT_)
	#define	_SM_INT_
#endif


/****************************************************************************************************/
/*                                                                                                  */
/* Constant definitions                                                                             */
/*                                                                                                  */
/****************************************************************************************************/

#define	SM_TYPE_POLYGON		0
#define	SM_TYPE_SPHERE		11
#define	SM_TYPE_DISK		12
#define	SM_TYPE_PARTDISK	13
#define	SM_TYPE_CYLINDER	14

#define	SM_PORT_VRP		0
#define SM_PORT_VPN		1
#define	SM_PORT_VUP		2
#define	SM_PORT_PRP		3
#define	SM_PORT_PICK		7

#define SM_PORT_PARALLEL	0
#define SM_PORT_PERSPECTIVE	1

#define	SM_OBJECT_FWD		10
#define	SM_OBJECT_UP		11
#define	SM_OBJECT_POS		12
#define	SM_OBJECT_EMISSION	13
#define	SM_OBJECT_AMBIENT	14
#define	SM_OBJECT_DIFFUSE	15
#define	SM_OBJECT_SPECULAR	16
#define	SM_OBJECT_BF_EMISSION	17
#define	SM_OBJECT_BF_AMBIENT	18
#define	SM_OBJECT_BF_DIFFUSE	19
#define	SM_OBJECT_BF_SPECULAR	20

#define	SM_OBJECT_SHADING_FLAT		0
#define	SM_OBJECT_SHADING_SMOOTH	1

#define	SM_OBJECT_UPDATE_NORMAL		1
#define	SM_OBJECT_UPDATE_POSITION	2
#define	SM_OBJECT_UPDATE_AMBIENT	4
#define	SM_OBJECT_UPDATE_DIFFUSE	8
#define	SM_OBJECT_UPDATE_EMISSION	16
#define	SM_OBJECT_UPDATE_SPECULAR	32
#define	SM_OBJECT_UPDATE_BF_AMBIENT	64
#define	SM_OBJECT_UPDATE_BF_DIFFUSE	128
#define	SM_OBJECT_UPDATE_BF_EMISSION	256
#define	SM_OBJECT_UPDATE_BF_SPECULAR	512

#define	SM_VERTEX_NORMAL_DEFAULT	0
#define	SM_VERTEX_NORMAL_USER		1

#define	SM_LIGHT_SOURCE_0		0
#define	SM_LIGHT_SOURCE_1		1
#define	SM_LIGHT_SOURCE_2		2
#define	SM_LIGHT_SOURCE_3		3
#define	SM_LIGHT_SOURCE_4		4
#define	SM_LIGHT_SOURCE_5		5
#define	SM_LIGHT_SOURCE_6		6
#define	SM_LIGHT_SOURCE_7		7

#define	SM_DEPTHCUE_EXP			0
#define	SM_DEPTHCUE_EXP2		1
#define	SM_DEPTHCUE_LINEAR		2


/****************************************************************************************************/
/*                                                                                                  */
/* Types and Structure definitions                                                                  */
/*                                                                                                  */
/****************************************************************************************************/


#ifdef USE_DOUBLE
typedef double DOUBLE;
#else
typedef float DOUBLE;
#endif


typedef struct DepthCue {
    int		mode;			/* depth-cue mode */
    XColor	*color;			/* depth-cue background color */
    DOUBLE	start, end, density;	/* depth-cue parameters */
} DepthCue;


typedef struct SpecularReflection {
    XColor	*color;			/* specular color */
    double	exponent;		/* specular exponent */
} SpecularReflection;


typedef struct LightSource {
    int		on;			/* whether this light source exists */
    XColor	*ambientColor;		/* ambient color */
    XColor	*diffuseColor;		/* diffuse color */
    XColor	*specularColor;		/* specular color */
    float	position[4];		/* position of light */
    float	direction[3];		/* direction of spotlight */
    float	ndirection[3];		/* normalized direction of spotlight */
    float	exponent;		/* spotlight exponent */
    float	cutoff;			/* spotlight cut-off */
    float	attenuation[3];		/* attenuation factors */
} LightSource;


typedef struct Normal3D {		/* three-dimensional normal vector */
    DOUBLE	v[3];
} Normal3D;


typedef struct Vertex {
    DOUBLE		lv[3];		/* local coordinates */
    struct Normal3D	lvn;		/* vertex normal in local coordinates system */
} Vertex;


typedef struct Surface {
    int			id;		/* surface id */
    int			vcnt;		/* number of vertices */
    int			*index;		/* pointer to array of vertex indices */
    int			backface;	/* backface flag */
    int			normal;		/* vertex normal flag */

    DOUBLE		ln[3];		/* surface normal in local coordinates system */
    struct Object	*obj;		/* object that owns this surface */
    struct Surface	*next;		/* pointer to next surface */
    struct Normal3D	*lvn;		/* pointer to array of vertex normals in local coordinates system */
} Surface;


typedef struct Object {
    struct Object	*next;			/* pointer to next object */
    struct Object	*parent;		/* pointer to parent object */
    struct Object	*child;			/* pointer to a child object */
    struct Object	*sibling;		/* pointer to a sibling object */
    struct Cell		*cell;			/* pointer to model cell */

    GLuint		pickname;		/* name assigned to the object during picking */

    int			type;			/* object type */
    int			updateFlag;		/* misc. update flags */

    char		*label;	        	/* name of the object */
    int			invalid;		/* whether the object's geometric specs are valid */
    int			shading;		/* shading method */
    int			ccw;			/* counter-clockwise front-facing mode */

    int			visible;		/* whether the object should be visible */
    int			useAverage;		/* whether vertex normals should be based on surface normal averages */
    GLenum		cull;			/* whether backface culling should be enabled */
    GLenum		frontmode;		/* drawing mode for front-facing polygons */
    GLenum		backmode;		/* drawing mode for back-facing polygons */

    GLUquadricObj	*quadric;
    GLenum		quadNormals;		/* desired normals */
    GLenum		orientation;		/* desired inside/outside orientation */
    GLenum		drawStyle;		/* desired draw style */
    double		radius;
    double		baseRadius;
    double		topRadius;
    double		innerRadius;
    double		outerRadius;
    double		startAngle;
    double		sweepAngle;
    double		height;
    int			slices;
    int			loops;
    int			stacks;

    char		*normals;		/* user-specified vertex normals */

    int			nv;			/* number of vertices */
    int			ns;			/* number of surfaces */
    Vertex		*v;			/* pointer to array of vertices */
    Surface		*s;			/* pointer to array of surfaces */

    DOUBLE		fwd[3];			/* forward vector */
    DOUBLE		up[3];			/* up vector */
    DOUBLE		pos[3];			/* position vector */
    DOUBLE		modelT[16];		/* model transform matrix */

    XColor		*color;			/* object's frontface color */
    XColor		*bfcolor;		/* object's backface color */

    DOUBLE		emission[3];		/* emission coefficients */
    DOUBLE		ambient[3];		/* ambience coefficients */
    DOUBLE		diffuse[3];		/* diffuse coefficients */
    SpecularReflection	specular;		/* specular parameters */

    float		emissionRGBA[4];	/* color of object modulated by emission coefficients */
    float		ambientRGBA[4];		/* color of object modulated by ambient coefficients */
    float		diffuseRGBA[4];		/* color of object modulated by diffuse coefficients */
    float		specularRGBA[4];	/* color of object modulated by specular coefficients */

    DOUBLE		bfemission[3];		/* backface emission coefficients */
    DOUBLE		bfambient[3];		/* backface ambience coefficients */
    DOUBLE		bfdiffuse[3];		/* backface diffuse coefficients */
    SpecularReflection	bfspecular;		/* backface specular parameters */

    float		bfemissionRGBA[4];	/* color of object modulated by backface emission coefficients */
    float		bfambientRGBA[4];	/* color of object modulated by backface ambient coefficients */
    float		bfdiffuseRGBA[4];	/* color of object modulated by backface diffuse coefficients */
    float		bfspecularRGBA[4];	/* color of object modulated by backface specular coefficients */

    Tcl_HashEntry	*e;			/* this object's entry in the cell's hash table */
} Object;


typedef struct PortItem  {
    Tk_Item		header;			/* Generic stuff that's the same for all
						 * types.  MUST BE FIRST IN STRUCTURE. */
    double		bbox[4];		/* Coordinates of bounding box for viewport
						 * (x1, y1, x2, y2).  Item includes
						 * x1 and x2 but not y1 and y2. */

    int			w, h;			/* viewport width and height */
    int			depth;			/* screen depth */

    int			srcx, srcy;		/* area of viewport actually drawn */
    int			srcw, srch;

    Tk_Window		tkwin;			/* the window that contains the MESA/OpenGL widget */
    Tk_Canvas		canvas;			/* the canvas widget that owns this item */
    Display		*dpy;			/* X display */
    XVisualInfo		*visinfo;		/* Visual Info record */
    Tcl_Command		widgetCmd;		/* Token for viewport's widget command */
    Pixmap		pm;			/* pixmap to draw into */
    GC			pmGC;			/* graphics context for image buffer */
    GLXPixmap		glxpm;			/* Mesa/OpenGL pixmap associated with the X pixmap */
    GLXContext		glxpmGC;		/* graphics context for Mesa/OpenGL */
    XColor		*outlineColor;		/* outline color */
    XColor		*fillColor;		/* fill color */
    GC			outlineGC;		/* graphics context for outline */

    int			initialized;		/* whether this port has been initialized */
    int			redraw;			/* whether image needs to be redrawn */
    int			update;			/* whether image update has been scheduled */
    int			invalid;		/* whether one or more viewport parameters have invalid values */
    int			hsr;			/* whether hidden surface removal should be performed */
    int			width;			/* width of outline. */

    Tcl_Interp		*interp;		/* the TCL interpreter in which this canvas is created */
    double		epsilon;		/* used in computing transformation vectors and in backface culling */

    int			projection;		/* current projection type (0: parallel, 1: perspective) */
    DOUBLE		vrp[3];			/* view reference point (WC) */
    DOUBLE		vpn[3];			/* view plane normal (WC) */
    DOUBLE		vup[3];			/* view up vector (WC) */
    DOUBLE		prp[3];			/* projection reference point (VRC) */

    DOUBLE		mr[16];			/* rotation matrix */
    DOUBLE		msh[16];		/* shear matrix */

    double		minu, maxu;		/* width of projection window */
    double		minv, maxv;		/* height of projection window */

    double		bp;			/* backplane (z) for clipping */

    float		ambient[4];		/* ambient RGBA intensities */
    XColor		*ambientLight;		/* ambient light */
    LightSource		lights[8];		/* light sources */

    DepthCue		depthcue;		/* depth cue parameters */

    int			pick[5];		/* pick[0] == 1 */
						/* coordinates (x1, y1) and (x2, y2) describing the pick region */

    int			invertedVisibleMask;	/* whether visible-masking should be reversed */
    char		*visibleMaskStr;	/* configuration value for visibleMask */
    Tcl_HashTable	visibleHash;		/* hash table for looking up visible mask values */

    struct Cell		*cell;			/* pointer to viewport's cell */
    struct PortItem	*next;			/* ointer to next viewport in the cell */
} PortItem;


typedef struct Cell {
    char		*label;			/* cell name */
    Object		*objects;		/* list of objects */
    PortItem		*ports;			/* list of viewports */
    struct Cell		*next;			/* pointer to next cell */
    Tcl_HashTable	hash;			/* hash table for looking up objects */
    Tcl_Interp		*interp;		/* cell's TCL interpreter */
    double		epsilon;		/* used in computing surface normals */
} Cell;


typedef struct CellHeader {
    Tcl_HashTable	hash;
    Cell		*cells;
} CellHeader;
